/*
 * Copyright (C) 2010 Apple Inc. All rights reserved.
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "OSAllocator.h"

#include "PageAllocation.h"
#include <errno.h>
#include <wtf/Assertions.h>
#include <wtf/UnusedParam.h>

#if PLATFORM(WIIU)
#include <PlatformOS.h>
#include <PlatformMemory.h>
#include <PlatformDefines.h>
using namespace LibPlatform;
#endif

//# define _USE_DLMALLOC
# define _USE_CODEGENAPI

// ----------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------
// - Low-Level-Allocator ------------------------------------------------------------------
// ----------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------

#if ENABLE(JIT) && defined(_USE_DLMALLOC)

class JITmemHandlerBase
{
public:
	JITmemHandlerBase() {};
	~JITmemHandlerBase() {};
	
	void init(void *pBase, unsigned size);
	
	void *alloc(unsigned size, unsigned align=0);
	void free(void *pAddr);
	
	unsigned pageSizeGet(void) const;
	void DbgAllocationError(void * /*pData1*/, void * /*pData2*/, void * /*pData3*/, void * /*pData4*/) {}
	
private:
	unsigned	m_hwPageSize;		//!< HW page size
	
	void *		m_pBase;			//!< Page size aligned base address of the memory managed by this instance
	unsigned	m_size;				//!< Page size aligned size of the memory managed by this instance
	
	void *		m_pNextAddr;		//!< Next address returned by pageAlloc()
	void *		m_pNextPage;		//!< Next page base address to be used by pageAlloc()
	unsigned	m_usedSize;			//!< Number of bytes used so far (from the base address of the memory managed)
	
	void *		m_pDLallocState;
	void *		m_pDLallocStateBase;
	unsigned	m_DLallocStateSize;
	
	OSMutex		m_mutex;
	
	void *pageAlloc(s32 bytes);

	static void *sbrk(JITmemHandlerBase *pHandler, s32 numBytes)		{ return(pHandler->pageAlloc(numBytes)); }
};

#include "wiiu/jit_malloc285.h"
namespace JITmalloc
{
  #include "wiiu/jit_malloc285.cpp"
};	// namespace JITmalloc

// --------------------------------------------------------------------------------
/**
*/
void JITmemHandlerBase::init(void *pBase, unsigned size)
 {
	OS::Report("JITmemHandlerBase: Init heap (%08x): %08x - %08x\n", this, pBase, (unsigned)pBase + size);
	
	m_hwPageSize = OSGetPageSize();

	m_pBase = (void*)(((unsigned)pBase + m_hwPageSize - 1) & ~(m_hwPageSize -1));
	m_size = size - ((unsigned)m_pBase - (unsigned)pBase);
	
	OS::Report("JITmemHandlerBase: Aligned heap (%08x): %08x - %08x\n", this, m_pBase, (unsigned)m_pBase + m_size);

	m_pNextPage = (void*)((unsigned)m_pBase + m_hwPageSize);
	m_pNextAddr = m_pNextPage;
	m_usedSize = m_hwPageSize;

	m_DLallocStateSize = m_hwPageSize;
	m_pDLallocStateBase = pBase;
	m_pDLallocState = JITmalloc::create_mspace_with_base(m_pDLallocStateBase, sbrk, this, m_DLallocStateSize, 0);
	
	OSInitMutex(&m_mutex);
}

// --------------------------------------------------------------------------------
/**
*/
void *JITmemHandlerBase::pageAlloc(s32 bytes)
{
	// Report mapped pool top address?
	if (bytes == 0)
		return(m_pNextPage);
		
	if (bytes > 0)
	{
		// Enlarge mapped pool...

		if (m_usedSize + bytes > m_size)
			return((void*)~0);

		void *pAddr = m_pNextAddr;

		bytes = (bytes + m_hwPageSize - 1) & ~(m_hwPageSize - 1);

		m_pNextAddr = (void*)((unsigned)m_pNextAddr + bytes);
		m_pNextPage = m_pNextAddr;
		
		m_usedSize += bytes;
		
		return pAddr;
	}
	
	// Shrink mapped pool...
// [...] don't do that, yet (would return new top address if it would shrink)
	return((void*)~0);
}

// --------------------------------------------------------------------------------
/**
*/
void *JITmemHandlerBase::alloc(unsigned size, unsigned align)
{
	OSLockMutex(&m_mutex);
	
	size_t allocatedSize;
	void *pAddr;
	if (align == 0)
		pAddr = JITmalloc::mspace_malloc(m_pDLallocState, size, &allocatedSize);
	else
		pAddr = JITmalloc::mspace_memalign(m_pDLallocState, align, size, &allocatedSize);

	OSUnlockMutex(&m_mutex);
	return(pAddr);
}

// --------------------------------------------------------------------------------
/**
*/
void JITmemHandlerBase::free(void *pAddr)
{
	OSLockMutex(&m_mutex);

	size_t	allocatedSize;
	JITmalloc::mspace_free(m_pDLallocState, pAddr, &allocatedSize);

	OSUnlockMutex(&m_mutex);
}

// --------------------------------------------------------------------------------
/**
*/
unsigned JITmemHandlerBase::pageSizeGet(void) const
{
	return(m_hwPageSize);
}

#endif

// ----------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------
// - WTF Support Functions for JIT allocations --------------------------------------------
// ----------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------

namespace WTF {

// ----------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------

#if ENABLE(JIT)
class WIIUcodeAllocator
{
public:
	WIIUcodeAllocator();
	~WIIUcodeAllocator();
	
	void	init(void);
	void	reset(void);

	void*	allocate(unsigned bytes);
	bool	free(void *address);
	
private:
	bool inited;
	void *					mpStart;
	void *					mpEnd;

	#if defined(_USE_DLMALLOC)
		JITmemHandlerBase		mHeapManager;
	#else
		Memory::HeapHandle			mHeap;
		OS::DynLoad_ModuleHandle	mModule;
	#endif
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

static WIIUcodeAllocator	_wiiuCodeAllocator;

// ----------------------------------------------------------------------------------------
/*
	Initialize wiiu code allocator for use
	
	@note	This is likely a TEMPORARY function, as loading a RPL will not be needed once
			a proper API is provided to access executable memory
*/
void wiiuJIT_StartupMemory(void)
{
	_wiiuCodeAllocator.init();
}

void wiiuJIT_StopMemory(void)
{
	_wiiuCodeAllocator.reset();
}
	
// ----------------------------------------------------------------------------------------
/*
	Initialize wiiu code allocator for use
	
	@note	This is likely a TEMPORARY function, as loading a RPL will not be needed once
			a proper API is provided to access executable memory
*/
void WIIUcodeAllocator::init(void)	
{
	#ifdef _USE_CODEGENAPI
		unsigned	base;
		size_t	size;
		OS::GetCodegenVirtAddrRange(&base, &size);
		
		if (size == 0)
		{
			OS::Report("*** Could not get codegen virtual address range. Is it disabled in cos.xml?\n");
			CRASH();
		}

		mpStart = (void*)base;
		mpEnd = (void*)(base + size);
	#else
		// Try loading the RPL containing the empty code block we use for JIT code...
		if (OS::DynLoad_Acquire("jit_memory", &mModule) != SYSTEM_DYNLOAD_ERR_NONE)
		{
			OS::Report("*** Could not load jit_memory.rpl! Did You copy it into the RPX' directory?\n");
			CRASH();
		}
		// That worked. Now grab the two symbols it exports to define our working area...
		OS::DynLoad_FindExport(mModule, 0, "JITMem_Start", &mpStart);
		OS::DynLoad_FindExport(mModule, 0, "JITMem_End", &mpEnd);
	#endif

//	OS::Report("JIT memory pool: %08x - %08x (core=%d)\n", mpStart, mpEnd, OS::GetCoreId());
	
	#if 0 //ndef NDEBUG
		OS::Report("JIT memory pool: Test write...\n");
		*(volatile unsigned*)mpStart = 0x01020304;
		OS::Report("JIT memory pool: Test write -> Done!\n");
		OS::Report("JIT memory pool: Test DCBST/SYNC/ICBI/ISYNC execution...\n");
        asm volatile (
			"dcbst	r0, %0\n"
			"sync\n"
			"icbi	r0, %0\n"
			"sync\n"
			".long 0x4C00012C	# isync\n"
            :					// out
            : "r" (mpStart)		// in
            : );				// clobber
		OS::Report("JIT memory pool: Test DCBST/SYNC/ICBI/ISYNC execution -> Done!\n");
	#endif
		
	// Create a heap in this area
	// (note: one should investigate if using a simpler heap manager might work out)
	#if defined(_USE_DLMALLOC)
 		mHeapManager.init(mpStart, size);
	#else
		mHeap = Memory::CreateExpHeap(mpStart, (unsigned)mpEnd - (unsigned)mpStart);
	#endif

	#ifdef _USE_CODEGENAPI
		// Lock down writes to area. Make it executable...
		OS::SwitchSecCodeGenMode(true);
	#endif
		inited = true;
}

// ----------------------------------------------------------------------------------------
void WIIUcodeAllocator::reset(void)
{
	if(inited)
	{
#ifdef _USE_CODEGENAPI
		// Allow writes to area...
		OS::SwitchSecCodeGenMode(false);
#endif
#if !defined(_USE_DLMALLOC)
		Memory::DestroyExpHeap(mHeap);
#endif
		inited = false;
		mpStart = 0;
		mpEnd = 0;
	}
}

// ----------------------------------------------------------------------------------------
/*
	Constructor for JIT memory manager
*/
WIIUcodeAllocator::WIIUcodeAllocator()
#if !defined(_USE_DLMALLOC)
: mHeap(0)
#endif
{
}

// ----------------------------------------------------------------------------------------
/*
	Destructor for JIT memory manager
*/
WIIUcodeAllocator::~WIIUcodeAllocator()
{
	reset();
}

// ----------------------------------------------------------------------------------------
/*
	Allocate from JIT pool
	
	@param	bytes	Number of bytes to allocate

	@return			Address of allocated memory block or NULL if not successful
	
*/
void* WIIUcodeAllocator::allocate(unsigned bytes)
{
	#ifdef _USE_CODEGENAPI
		// Allow writes
		OS::SwitchSecCodeGenMode(false);
	#endif
	#if defined(_USE_DLMALLOC)
		return mHeapManager.alloc(bytes);
	#else
		ASSERT(mHeap);
		return Memory::AllocFromExpHeapEx(mHeap, bytes, pageSize());
	#endif
	#ifdef _USE_CODEGENAPI
		// Allow execute
		OS::SwitchSecCodeGenMode(true);
	#endif
}

// ----------------------------------------------------------------------------------------
/*
	Free to JIT pool
	
	@param	address	Address of memory block to be released
	
	@return			True if the address was managed in the JIT pool  false otherwise
*/
bool WIIUcodeAllocator::free(void *address)
{
	// Our memory?
	if (mpStart > address || mpEnd <= address)
		return false;	// No...
		
	// Yes, free it...
	#ifdef _USE_CODEGENAPI
		// Allow writes
		OS::SwitchSecCodeGenMode(false);
	#endif
	#if defined(_USE_DLMALLOC)
		mHeapManager.free(address);
	#else
		ASSERT(mHeap);
		Memory::FreeToExpHeap(mHeap, address);
	#endif
	#ifdef _USE_CODEGENAPI
		// Allow execute
		OS::SwitchSecCodeGenMode(true);
	#endif
	return true;
}

#endif

// ----------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------

void* OSAllocator::reserveUncommitted(size_t bytes, Usage usage, bool writable, bool executable, bool includesGuardPages)
{
    void* result = reserveAndCommit(bytes, usage, writable, executable, includesGuardPages);
    return result;
}

void* OSAllocator::reserveAndCommit(size_t bytes, Usage usage, bool writable, bool executable, bool includesGuardPages)
{
    void* result = 0;

#if ENABLE(JIT)
	// If the memory needs to be executable, we need to ask the JIT pool for it...
	if (executable)
		return(_wiiuCodeAllocator.allocate(bytes));
#endif

	result = Memory::AllocFromUsageHeap (usage, bytes, pageSize());
    if (result == NULL) {
#if ENABLE(INTERPRETER)
        if (executable)
            result = 0;
        else {
#endif
        	Memory::HeapHandle baseHeapMEM2;
            baseHeapMEM2 = Memory::GetBaseHeapHandleArena2();
            Memory::DumpHeap(baseHeapMEM2); // for debugging only.
            CRASH();
#if ENABLE(INTERPRETER)
			}
#endif
    }

    return result;
}

void OSAllocator::commit(void* address, size_t bytes, bool, bool)
{
    UNUSED_PARAM(address);
    UNUSED_PARAM(bytes);
}

void OSAllocator::decommit(void* address, size_t bytes)
{
    UNUSED_PARAM(address);
    UNUSED_PARAM(bytes);
}

void OSAllocator::releaseDecommitted(void* address, size_t bytes)
{
#if ENABLE(JIT)
	// Check if the JIT pool manages the memory and then free it either there or in the default heap
	if (!_wiiuCodeAllocator.free(address))
	{
		Memory::FreeToUsageHeap(address);//DefaultHeap(address);
	}
#else
	Memory::FreeToUsageHeap(address);
#endif
}

} // namespace WTF
