/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef GRAPHICS_CONTEXT_PLATFORM_PRIVATE_H_
#define GRAPHICS_CONTEXT_PLATFORM_PRIVATE_H_

#include "config.h"
#include "GraphicsContext.h"
#include "ImageSource.h"
#include "ImageBuffer.h"
#include "Color.h"
#include "IntRect.h"
#include "wtf/Vector.h"
#include "wtf/ByteArray.h"
#include "ShadowBlur.h"
#include "TimeRanges.h"

#include <vector>

// -------------------------------------------------------------------------------------
// Platform specifics
//

#if !PLATFORM(WIIU)
#if OS(WINDOWS)
#include <windows.h>
#else
#include <pthread.h>
#endif

#include "WiiURenderParams.h"
#include "renderer/WiiURenderer.h"
#endif

// -------------------------------------------------------------------------------------
// WKgfx System includes
//
#include <wkgfx_Matrix.h>
#include <wkgfx_Misc.h>

// -------------------------------------------------------------------------------------
// Forward definitions
//

namespace WebCore {
	class FloatSize;
	class Color;
	class SimpleFontData;
	class GlyphBuffer;
	class ImageBufferData;
	class PathPolygon;
}

namespace WTF {
	class StatisticsWiiU;
}

namespace WKgfx {
	struct TransparencyLayer;
	class GCState;
	class ClipOp;
	struct Private;
}

// -------------------------------------------------------------------------------------

//! Platform specific graphics context
struct GraphicsContextPlatformPrivate {

	enum FillPolygonMode {
		FILL_POLYGONS_CLIP,				//!< Render path polygons, adding to the clip mask
		FILL_POLYGONS_CLIPOUT,			//!< Render path polygons, taking away from the clip mask
		FILL_POLYGONS_FILL				//!< Render path polygons to color buffer
	};
	
	enum SHADER {
		SHADER_COLOR = 0,				//!< Simple colored shader
		SHADER_TEXTURE,					//!< Simple alpha-tinted texture shader
		SHADER_TEXTURE_CALC_VTX,		//!< Same as SHADER_TEXTURE, but intended to calculate vertex data in the shader from normalized position data
		SHADER_TEXTURE_CALC_VTX_NO_PREMULTIPLY,		//!< Same as SHADER_TEXTURE, but intended to calculate vertex data in the shader from normalized position data
		SHADER_FONT,					//!< Font shader (gray scale texture is input as alpha, color is constant per tint)
		SHADER_COPY,					//!< Shader to copy data
		SHADER_COPY_ALPHAMUL,			//!< Shader to copy data and pre-multiply alpha in source
		SHADER_PATTERN,					//!< Shader to render a textured pattern with specialized texture fetches to allow pattern-style rendering (UVs are synthesized)
		SHADER_PATTERN_NO_PREMULTIPLY,	//!< Shader to render a textured pattern with specialized texture fetches to allow pattern-style rendering (UVs are synthesized)
		SHADER_PATTERN_SIMPLE,			//!< Shader just like SHADER_PATTERN, but with just simple texture fetches (fast, for use with "full coverage" image use)
		SHADER_PATTERN_SIMPLE_NO_PREMULTIPLY,	//!< Shader just like SHADER_PATTERN, but with just simple texture fetches (fast, for use with "full coverage" image use)
		SHADER_YUV,						//!< Render YUV 422 data (in two textures) - otherwise: like SHADER_TEXTURE
		SHADER_GAUSSIAN_BLUR,			//!< Apply gaussian blur pass (configurable tabs etc.)
		
		SHADER_NUM,
		
		SHADER_NONE = 0xffff			//!< Shader is not yet set
		};
	
	enum SAMPLER {
		SAMPLER_POINT = 0,
		SAMPLER_POINT_WRAPST,
		SAMPLER_BILINEAR,
		SAMPLER_BILINEAR_WRAPS,
		SAMPLER_BILINEAR_WRAPT,
		SAMPLER_BILINEAR_WRAPST,
		
		SAMPLER_NUM,
		
		SAMPLER_NONE = 0xffff
		};
		
	enum VTXTYPE {
		VTX_POSITION = 0,				//!< 2D position-only vertex (all float)
		VTX_TEXTURED,					//!< 2D position & 2D-texture coordinates vertex (all float)
		};
	
	enum STENCIL_CLEAR {
		STENCIL_CLEAR_EMPTY = 0,		//!< Setup stencil to "everything clipped" state
		STENCIL_CLEAR_FILLED			//!< Setup stencil to "nothing clipped" state
	};
	
	enum TextureFormat {
		TEX_FMT_R8 = 0,					//!< Simple 8-bit gray scale texture (delivered in red component)
		TEX_FMT_RGBA8,					//!< Simple 8-bit RGBA texture
		TEX_FMT_RGB565,					//!< RGB565 16-bit texture (no alpha)
		TEX_FMT_RGBA4444,				//!< RGBA4444 16-bit texture
		TEX_FMT_RGBA5551,				//!< RGBA5551 16-bit texture, 1-bit alpha
		TEX_FMT_BC1,					//!< Block Compressed format
	};
	
	// ------------------------------------------------------------------

	//
	// Management / Core
	//

	GraphicsContextPlatformPrivate();
	virtual ~GraphicsContextPlatformPrivate();

	void setSize(int width, int height);
	bool init(WebCore::GraphicsContext *pOwner, int aaMode = 0, bool useMem1=false, int disableStencil=-1, int format=0);
	
	int	width() const									{ return m_width; }
	int	height() const									{ return m_height; }
	
	static void startup(WKgfx::CBpostDLCall postDLCall);
	static void shutdown(void);
	
	void *prepareForPaint(bool bSync = false);
	static void prepareForProcessSwitch(bool bForeground);

	//
	// State
	//

	bool makeCurrent();

	void pushState();
	void popState();
	
	//
	// Transform
	//

	void setViewMtx(const WKgfx::Matrix & mat);
	void concatViewMtx(const WKgfx::Matrix & mat);
	const WKgfx::Matrix & getViewMatrix(void) const;
	const WKgfx::Matrix & getProjMatrix(void) const;	

	void setViewMtxLoadPointer(const WKgfx::Matrix *pMtx) { m_pCurrentViewMtx = pMtx; }

	//
	// General
	//

	void setAlpha(float alpha);
	float getAlpha() const;
	//
	// Strokes
	//

	WebCore::LineCap getLineCap();
	WebCore::LineJoin getLineJoin();

	void setLineCap(WebCore::LineCap lineCap);
	void setLineJoin(WebCore::LineJoin lineJoin);
	void setMiterLimit(float miter);
	void setStrokeThickness(float strokeThickness);
	
	//
	// Clipping
	//
	
	void setDirtyRect(const WebCore::IntRect & contentRect);

	void resetClipOps();
	void addClipOp(const WKgfx::ClipOp & clip_op);
	void resetClipRect();   
	void intersectClipRect(const WebCore::IntRect & irect);

	WebCore::IntRect clipBounds() const;
	void clipRect(const WebCore::FloatRect& rect);
	
	//
	// Shadows
	//
	
	void enableShadow(WebCore::FloatSize const& size, float blur, WebCore::Color const& color);
	void enableShadow();
	void disableShadow();
	
	bool isShadowed() const;
	float shadowWidth() const;
	float shadowHeight() const;
	const WebCore::Color & shadowColor() const;

    WebCore::ShadowBlur & getShadowBlurHandler()	{ return m_shadowBlurHandler; }

	//
	// Compose
	//

	void setCompositeOperation(WebCore::CompositeOperator op);

	void beginPlatformTransparencyLayer(float opacity);
	void endPlatformTransparencyLayer();
	
	//
	// Image Filters
	//

	bool imageFilterIsEnabled() const;
	void setImageInterpolationQuality(WebCore::InterpolationQuality quality);
	
	//
	// Textures
	//
	
	WebCore::NativeImagePtr createTextureDataFromContext(void);
	WebCore::NativeImagePtr getTexture() const;
	int getTextureDataSize() const;
	
	static WebCore::NativeImagePtr createTextureFromMemory(int width, int height, unsigned int *m_bytes, TextureFormat fmt=TEX_FMT_RGBA8, bool dither=false);
	static WebCore::NativeImagePtr createTextureFromMemoryOnly(int width, int height, unsigned int *m_bytes, TextureFormat fmt=TEX_FMT_RGBA8);	// ONLY creates the texture, doesn't copy it!
	static void convertTextureFromMemoryToNative(WebCore::NativeImagePtr, int width, int height, void *m_bytes, bool dither=true);	// this does the actual copy!
	static void convertTextureFromMemoryToNativeR8(WebCore::NativeImagePtr, int width, int height, void *m_bytes);	// this does the actual copy (RGB565 format!)
	static void convertTextureFromMemoryToNative565(WebCore::NativeImagePtr, int width, int height, void *m_bytes, bool dither=true);	// this does the actual copy (RGB565 format!)
	static void convertTextureFromMemoryToNative5551(WebCore::NativeImagePtr, int width, int height, void *m_bytes, bool dither=true);	// this does the actual copy (RGB565 format!)
	static void convertTextureFromMemoryToNative4444(WebCore::NativeImagePtr, int width, int height, void *m_bytes, bool dither=true);	// this does the actual copy (RGBA4444 format!)
	static void convertTextureFromMemoryToNativeBC1(WebCore::NativeImagePtr, int width, int height, void *m_bytes);	// this does the actual copy (BC1 format!)
	
	static void finishTextureConversion(WebCore::NativeImagePtr, bool doBegin=true);	// this does the actual copy!
	static void copyTextureFromMemoryToNative(WebCore::NativeImagePtr, int width, int height, unsigned int *m_bytes, bool dither=false);	// this does the actual copy!
	static WebCore::NativeImagePtr createTexture(int width, int height, TextureFormat fmt);
	static void destroyTextureData(WebCore::NativeImagePtr texture);
	static void destroyTextureData2(WebCore::NativeImagePtr texture);
	static void *accessTextureBuffer(WebCore::NativeImagePtr texture, int & width, int & height, int & pitch);
	static void flushCacheForTexture(WebCore::NativeImagePtr texture);
	static void flushTextureData(void *pAddr, int size);
	static void* allocatePixelData(int size);	// allocate a buffer from a special heap
    static void* allocateTextureData(int size, int align = 4);	// allocate a buffer from a special heap
	static int getTextureDataSize(WebCore::NativeImagePtr texture);
	static void freePixelData(void *dataP);
	
	//
	// Images
	//
	
	WTF::PassRefPtr<WTF::ByteArray> getUnmultipliedImageData(const WebCore::IntRect& rect, const WebCore::IntSize& size);
	WTF::PassRefPtr<WTF::ByteArray> getPremultipliedImageData(const WebCore::IntRect& rect, const WebCore::IntSize& size);
	void putUnmultipliedImageData(WTF::ByteArray* source, const WebCore::IntSize& sourceSize, const WebCore::IntRect& sourceRect, const WebCore::IntPoint& destPoint, WebCore::ImageBufferData& data, const WebCore::IntSize& size);
	void putPremultipliedImageData(WTF::ByteArray* source, const WebCore::IntSize& sourceSize, const WebCore::IntRect& sourceRect, const WebCore::IntPoint& destPoint, WebCore::ImageBufferData& data, const WebCore::IntSize& size);
	void transformColorSpace(const WTF::Vector<int>& lookUpTable, const WebCore::IntSize& size);
	void *getContextDataAsRGBA8(int & width, int & height, int & pitch);
	void *getContextDataAsRGBA8Stretched(int dstWidth, int dstHeight, int srcX = 0, int srcY = 0, int srcWidth = -1, int srcHeight = -1);

	void applyGaussianBlur(WebCore::GraphicsContext * pLayerContext, const WebCore::FloatPoint & layerOrigin, WebCore::GraphicsContext *pScratchContext, const WebCore::FloatSize & blurRadius);
	
	//
	// Rendering
	//

	void drawDataYUV(const WebCore::FloatRect & rect, const WebCore::FloatRect & src, void* pY, void* pUV, int width, int height);
	void drawImageStream(int channel, float x, float y, float width, float height);
	void drawImage(WebCore::NativeImagePtr texture, const WebCore::FloatRect& dst, const WebCore::FloatRect& src, WebCore::CompositeOperator op);
	void drawPattern(WebCore::NativeImagePtr texture, const WebCore::FloatRect& tileRect, const WebCore::FloatPoint& phase, const WebCore::FloatRect& destRect, const WebCore::AffineTransform& patternTransform);
	void drawPatternFast(WebCore::NativeImagePtr texture, const WebCore::FloatRect& destRect, bool premultipliedAlpha = true);
	void fillRect(const WebCore::FloatRect& rect, const WebCore::Color& color);
	void drawLineForText(const WebCore::FloatPoint& origin, float width, const WebCore::Color& color);
	void clearRect(const WebCore::FloatRect& rect);
	void strokeRect(const WebCore::FloatRect& rect, float width, const WebCore::Color& color);
	void drawRect(const WebCore::IntRect& rect, const WebCore::Color & color, bool bDoStroke, const WebCore::Color & strokeColor);
	void drawLine(const WebCore::IntPoint& point1, const WebCore::IntPoint& point2, WebCore::StrokeStyle style, const WebCore::Color & color);
	void drawEllipse(const WebCore::IntRect& rect, const WebCore::Color & color, const WebCore::Color & strokeColor, WebCore::StrokeStyle style);
	void strokeArc(const WebCore::IntRect& rect, int startAngle, int angleSpan, const WebCore::Color & color, WebCore::StrokeStyle style);
	void drawConvexPolygon(size_t npoints, const WebCore::FloatPoint* points, const WebCore::Color & fillColor, bool bDoStroke, const WebCore::Color & strokeColor);
	void fillPolygons(const WTF::Vector<WebCore::PathPolygon>& polygons, FillPolygonMode mode, const WebCore::AffineTransform* transformation, WebCore::WindRule fillRule, const WebCore::Color * color = NULL);
	void strokePolygons(const WTF::Vector<WebCore::PathPolygon>& polygons, const WebCore::AffineTransform* transformation, bool bClip, const WebCore::Color & color);
	void gradientFill(PlatformGradient pGradient, const WebCore::FloatRect& rect);
	void drawGlyphs(const WebCore::SimpleFontData* font, const WebCore::GlyphBuffer& glyphBuffer, int from, int numGlyphs, const WebCore::FloatPoint& point, const WebCore::Color & fillColor);

#if PLATFORM(WIIU)
	void strokeGlyph(const WebCore::SimpleFontData* font, const WebCore::GlyphBuffer& glyphBuffer, int from, int numGlyphs, const WebCore::FloatPoint& point, float strokeThickness, const WebCore::Color & fillColor);
#else
	bool strokeGlyph(WebCore::Path & p, const WebCore::GlyphBuffer& glyphBuffer, int glyphIndex, const WebCore::SimpleFontData* font); 
#endif
	//
	// Custom Rendering for MobiClip Media Player
	//
	void beginCustomDraw();
	void endCustomDraw(unsigned sizeHint);
	
	//
	// Extension Support
	//

	// TGM: Experimental canvas shaders
	unsigned long loadShader(const String& gsh);
	unsigned long loadTexture(const String& gtx);
	void setColorShader(unsigned long shaderId);
	void setTextureShader(unsigned long shaderId);
	void setFontShader(unsigned long shaderId);
	void setPatternShader(unsigned long shaderId);
	void setVertexUniformFloat(unsigned long reg, float x, float y, float z, float w);
	void setPixelUniformFloat(unsigned long reg, float x, float y, float z, float w);
	void setPixelUniformTexture(unsigned long reg, unsigned long textureId);
	unsigned long getColorShader(void);
	unsigned long getTextureShader(void);
	unsigned long getFontShader(void);
	unsigned long getPatternShader(void);
	void customShaderClearStateFlags(void);

	static void customShaderCleanup(void);

	// TGM: Image smoothing control
	void setEnableImageSmoothing(bool enable);
	bool getEnableImageSmoothing(void);

	// TGM: Native rendering
	void *getNativeColorBuffer(void);
	void *getNativeDepthBuffer(void);

	//
	// Low Level
	//
	
	void setShader(SHADER shader);
	void setSampler(unsigned int samplerIdx, SAMPLER sampler);
	void forceSetShader(SHADER shader);

	void enableStencilTest();
	void setStencilHardware(bool bTestEnable);
	void resetStencilBuffer(STENCIL_CLEAR type);
	
	void setScissorHardware(const WebCore::IntRect & rect);

	bool prepareForUseAsTexture();
	void invalidateBuffers();

#if OS(WINDOWS) || OS(UNIX)
	WiiURenderer *getRenderer(void);
	void setRenderer(WiiURenderer *set);
	WiiUWindowType getWindowType(void);
	void setWindowType(WiiUWindowType set);
	WiiURenderer::tRenderContext *getRenderContext(void);
	void SetMipmapGeneration(const bool vEnabled);
#endif

	static bool isDrawEnabled()						{ return s_bDrawEnabled; }
	static void setDrawEnabled(bool bEnable)		{ s_bDrawEnabled = bEnable; }

	// -------------------------------------------------------------------
	
private:
	friend class WKgfx::ClipOp;
	friend void clearDrawActive(void);

	
	bool								m_bInit;						//!< True if this instance is initialized
	bool								m_bInitialGPUSetup;				//!< True if the initial GPU setup (clearing of the render buffers & some state) has not been done, yet
	bool								m_bContentMaybeDirty;			//!< True after context has been made current at least once until invalidateBuffers was called at last once (signifies that the color & depth buffers may need flushing)
	
	WebCore::GraphicsContext *  		m_pOwner;						//!< Owning platform independent context
	static WebCore::GraphicsContext *   s_pLastToRender;				//!< Current context or NULL

	int									m_width;						//!< Width of context in pixels
	int									m_height;						//!< Height of context in pixels

	const WKgfx::Matrix *				m_pCurrentViewMtx;				//!< Pointer to currently loaded view matrix data
	WKgfx::GCState *					m_pState;						//!< Current graphics state shadow
	std::vector<WKgfx::GCState*>		m_savedStates;					//!< Saved state values

	std::vector<WKgfx::TransparencyLayer *>	m_Transparency_layers;		//!< Transparency layers currently in use

    WebCore::ShadowBlur					m_shadowBlurHandler;			//!< Shadow blur rendering handler

	SHADER								m_CurrentShader;				//!< Currently loaded shader
	SAMPLER								m_CurrentSampler[2];			//!< Current samplers

	WKgfx::Private *					m_pPrivate;						//!< Private platform data

	static const float					s_idMtx[16];					//!< Identity matrix for fast access


	void setState();
	void setCurrentShader();
	void setCurrentSampler(unsigned int samplerIdx);
	void setupCompositeOperation();

    void ChangeShaderToBeUsed(SHADER shaderToChange, void *pNewShader);
    void RestoreDefaultShader(SHADER shaderToChange);

	static void resourcesSetup();
	static void resourcesFinishSetup();
	static void resourcesCleanup();

	//
	// Platform Specific
	//

	static void _startup(WKgfx::CBpostDLCall postDLCall);
	static void _shutdown(void);

	void vtxBegin(unsigned int num, VTXTYPE type);
	void vtxPos(float x, float y);
	void vtxPosTex(float x, float y, float u, float v);
	void vtxEnd();
	
	void initialGPUSetup();
	static void _beginDraw();

	void drawTextureYUV(const WebCore::FloatRect& dst, const WebCore::FloatRect& src, const void *textureY, const void *textureUV, float fadeAmount = 1.0f);

	void pathSimplePathDrawFilled(WebCore::Path & path);
	
	void putImageData(WTF::ByteArray*& source, const WebCore::IntSize& sourceSize, const WebCore::IntRect& sourceRect, const WebCore::IntPoint& destPoint, WebCore::ImageBufferData& data, const WebCore::IntSize& size, WebCore::Multiply multiplied);
	template <WebCore::Multiply multiplied> WTF::PassRefPtr<WTF::ByteArray> getImageData(const WebCore::IntRect& rect, const WebCore::IntSize& size);

	void setPolygonMode(void)	{ if (m_bLineOffset)  { m_bLineOffset = false; setRasterOffset(false); } }
	void setLineMode(void)		{ if (!m_bLineOffset) { m_bLineOffset = true;  setRasterOffset(true);  } }
	void setRasterOffset(bool bEnable);
	
	static void setVertexUniform(unsigned uniform, const WebCore::Color & color, float alphaScale);
	
	void startCompositeOperation();
	void endCompositeOperation();
	void renderTexture(WebCore::NativeImagePtr texture, WebCore::FloatRect dst, WebCore::FloatRect src);

#if OS(WINDOWS) || OS(UNIX)
public:
#endif
	static bool beginDraw();
#if OS(WINDOWS) || OS(UNIX)
private:
#endif
	static void endDraw(bool bSync = false);
	static void waitForGPU();
	static bool isDrawActive()			{ return s_bDrawActive; }

	static void checkDrawStatus(unsigned sizeHint);
	
	static bool IsWebKitThread();

	bool								m_bLineOffset;					//!< If true uniforms are setup to provide +0.5 offset to screen coordinates (both axis), if false: offset is zero
	float								m_lineOff[4];					//!< Applicable 2D line offset for this context
	static const float					s_zeroOff[4];					//!< Zero 2D line offset value

	WKgfx::Matrix						m_projMtx;						//!< Projection matrix to be used with this instance

	float *								m_pVtxBase;						//!< Valid during vertex insertion: vertex buffer base
	float *								m_pVtx;							//!< Valid during vertex insertion: vertex buffer write pointer
	unsigned int						m_vtxNum;						//!< Valid during vertex insertion: size of vertex buffer in vertices
	unsigned int						m_vtxSize;						//!< Valid during vertex insertion: size of a single vertex in bytes
	bool								m_bVtxHasTex;					//!< Valid during vertex insertion: true if current vertex type is textured

	static bool							s_bDrawActive;					//!< True if draw environment (for all contexts) has been setup
	static bool							s_bDrawEnabled;					//!< True if rendering is allowed (false only during shutdown - afterwards state is not guaranteed anymore)

	bool								m_needsCompositing;				//!< True if the start and end compositing calls should be invoked, false otherwise.

	static void *						s_pWebkitThread;				//!< Current webkit thread (or NULL, if not yet known)
	#if !PLATFORM(WIIU) && !OS(WINDOWS)
	static bool                   s_pWebkitThreadSet;
	#endif

	struct InvalidateEntry
	{
		InvalidateEntry() {}
		InvalidateEntry(int invalidateType, void *pAddr, int size) : m_invalidateType(invalidateType), m_pAddr(pAddr), m_size(size) {}
	
		int		m_invalidateType;
		void *	m_pAddr;
		int		m_size;
	};

	static std::vector<InvalidateEntry>	*m_invalidateList;

	static void enqueueInvalidate(int invalidateType, void *pAddr, int size);
	static void processInvalidates(void);
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	class WKgfxHooks : public WKgfx::CallbackHooks 
	{
		virtual void ShadowBlur_ShadowBlur(WebCore::ShadowBlur *pThis);
		virtual void ShadowBlur_setShadowValues(WebCore::ShadowBlur *pThis, const WebCore::FloatSize&, const WebCore::FloatSize& , const WebCore::Color&, WebCore::ColorSpace, bool ignoreTransforms);
		virtual WTF::StatisticsWiiU *StatisticsWiiU_instance(void);
		virtual void Path_Path(WebCore::Path *pThis);
		virtual void Path_DestructPath(WebCore::Path *pThis);
		virtual void Path_addArc(WebCore::Path *pPath, const WebCore::FloatPoint&, float radius, float startAngle, float endAngle, bool anticlockwise);
		virtual void Path_moveTo(WebCore::Path *pPath, const WebCore::FloatPoint &);
		virtual void Path_addLineTo(WebCore::Path *pPath, const WebCore::FloatPoint &);
		virtual void Path_addQuadCurveTo(WebCore::Path *pPath, const WebCore::FloatPoint &, const WebCore::FloatPoint &);
		virtual void Path_addBezierCurveTo(WebCore::Path *pPath, const WebCore::FloatPoint &, const WebCore::FloatPoint &, const WebCore::FloatPoint &);
		virtual WebCore::FloatPoint AffineTransform_mapPoint(const WebCore::AffineTransform *pThis, const WebCore::FloatPoint& point);
		virtual void IntRect_intersect(WebCore::IntRect *pThis, const WebCore::IntRect & rect);
		virtual void IntRect_IntRect(WebCore::IntRect *pThis, const WebCore::FloatRect & rect);
		virtual void FloatRect_fitToPoints(WebCore::FloatRect *pThis, const WebCore::FloatPoint& p0, const WebCore::FloatPoint& p1, const WebCore::FloatPoint& p2, const WebCore::FloatPoint& p3);
		virtual void FloatRect_FloatRect(WebCore::FloatRect *pThis, const WebCore::IntRect & rect);
		virtual WTF::PassRefPtr<WTF::ByteArray> ByteArray_create(size_t size);
		virtual WTF::CString String_ascii(const WTF::String *pThis);
		virtual WTF::CString String_utf8(const WTF::String *pThis);
		virtual bool Mutex_tryLock(Mutex *pThis);
		virtual void Mutex_unlock(Mutex *pThis);
		virtual WTF::ThreadIdentifier WTF_currentThread(void);
		virtual bool WTF_isMainThread(void);
		virtual const WebCore::FontData* FontFallbackList_fontDataAt(const WebCore::FontFallbackList *pThis, const WebCore::Font* pFont, unsigned index);
		virtual WebCore::SimpleFontData* FontCache_getCachedFontData(WebCore::FontCache *pThis, const WebCore::FontPlatformData *pData, WebCore::FontCache::ShouldRetain bRetain);
		virtual WebCore::SimpleFontData* FontCache_getCachedFontData(WebCore::FontCache *pThis, const WebCore::FontDescription& font, const char* familyName, bool checkingAlternateName, WebCore::FontCache::ShouldRetain );
		virtual void FontPlatformData_FontPlatformData(WebCore::FontPlatformData *pThis, font_config_handle *pHandle, font_param *pParam, const WebCore::FontDescription & desc);
		virtual void FontPlatformData_FontPlatformData(WebCore::FontPlatformData *pThis, font_face_t *face, char * familyName, float size, bool bBold, bool bItalic);
		virtual void FontPlatformData_FontPlatformData(WebCore::FontPlatformData *pThis, const WebCore::FontPlatformData & data);
		virtual void FontPlatformData_DestructFontPlatformData(WebCore::FontPlatformData *pThis);
		virtual void WebKitCrash(void);
	};
	
	static WKgfxHooks *			s_pWKgfxHooks;
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
};

// --------------------------------------------------------------------------------------------------
// - Inlines ----------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------

// --------------------------------------------------------------------------------------------------
/**
*/
inline PlatformGraphicsContext *WebCore::GraphicsContext::platformContext() const
{ 
	return m_data;
}

#endif // GRAPHICS_CONTEXT_PLATFORM_PRIVATE_H_

