/*
 * Copyright (C) 2009 Brent Fulgham.  All rights reserved.
 * Copyright (C) 2009 Google Inc.  All rights reserved.
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "SocketStreamHandle.h"

#include "KURL.h"
#include "Logging.h"
#include "NotImplemented.h"
#include "SocketStreamHandleClient.h"
#include "SocketStreamHandlePrivate.h"
#include "SocketStreamError.h"
#include <stdio.h>
#include <errno.h>

#include "CString.h"
#include "MainThread.h"

#if PLATFORM(WIIU) || PLATFORM(WIIU_DEV)
#include "DomainAccessControlWiiU.h"
#include "PlatformNetwork.h"
#include "PlatformOS.h"
#include "PlatformError.h"
#include "PlatformSSL.h"
using namespace LibPlatform;
#endif

using namespace std;

namespace WebCore {

SocketStreamHandlePrivate::SocketStreamHandlePrivate(SocketStreamHandle* streamHandle)
: m_streamHandle(streamHandle)
, m_timer(this, &SocketStreamHandlePrivate::checkSocketTimerFired)
, setup_timer(this, &SocketStreamHandlePrivate::markOpen)
{

}

SocketStreamHandlePrivate::~SocketStreamHandlePrivate()
{
    // stop the timer
    m_timer.stop();
}

void SocketStreamHandlePrivate::readReady() {
	CURLcode rc;     // verify that we still have a valid handle before doing a read     if ( !m_streamHandle || !m_CURLhandle  ) {         return;     }      memset(buf, '\0', 8192);

    size_t iolen;
    rc = curl_easy_recv(m_CURLhandle, buf, 8192, &iolen);
    if ( iolen <= 0 ) {
        return;
    }
    if ( rc == CURLE_AGAIN ) {
        return;
    }
    if( rc == CURLE_OK ) {
        m_streamHandle->client()->didReceiveSocketStreamData(m_streamHandle, buf, iolen);
    }
}

int SocketStreamHandlePrivate::send(const char* buf, int length)
{
	size_t bytes_out = 0;
    CURLcode rc;
    int bytes_sent = 0;
    if (!m_streamHandle ) {
        return 0;
    }
    rc = curl_easy_send(m_CURLhandle, buf, length, &bytes_out);
    bytes_sent += bytes_out;
    if ( rc != CURLE_OK ) {
        LOG(Network, "Error: failed to write %d bytes to socket\n", length);
        return 0;
    }

    m_streamHandle->sendPendingData();

    return bytes_sent;
}

int SocketStreamHandlePrivate::getSocketInfo()
{
    CURLcode rc;
    // check to see if we're closing before grabbing the socket
    if (!m_CURLhandle ) {
        return 0;
    }

    rc = curl_easy_getinfo(m_CURLhandle,CURLINFO_LASTSOCKET, &m_socket);
    if ( rc != CURLE_OK ) {
        LOG(Network, "Failed to obtain socket from curl handle : %p\n", m_CURLhandle);
        return 0;
    }
    return 1;
}

void SocketStreamHandlePrivate::markOpen(Timer<SocketStreamHandlePrivate>*)
{
	m_streamHandle->client()->didOpenSocketStream(m_streamHandle);
}


void SocketStreamHandlePrivate::checkSocketTimerFired(Timer<SocketStreamHandlePrivate>*)
{
    fd_set infd;
    int res;
    FD_ZERO(&infd);
    FD_SET(m_socket, &infd);
    res = select(m_socket + 1, &infd, NULL, NULL, &timeout);
    if ( res == -1 ) {
        return;
    } else if ( res == 0 ) {
        return;
    }
    if ( FD_ISSET(m_socket, &infd) ) {
    	readReady();
    }
}
int SocketStreamHandlePrivate::initialize(const KURL& url)
{
    KURL urlcopy = url;
    CURLcode rc;
	unsigned int retryCount = 0;

#if PLATFORM(WIIU)
	int error = WebCore::DomainAccessControl::DOMAIN_ACCESS_OK;
	if (!WebCore::DomainAccessControl::IsDomainAccessible(url.string(), &error))
	{
		switch (error)
		{
		case WebCore::DomainAccessControl::DOMAIN_ACCESS_NO_NETWORK_ACCESS:
			OS::Report("This request cannot be honored. You must enable network access.\n\tRequested URL: %s.\n",
					 encodeWithURLEscapeSequences(url.string()).latin1().data());
			LOG(Network, "This request cannot be honored. You must enable network access.\n\tRequested URL: %s.\n",
				encodeWithURLEscapeSequences(url.string()).latin1().data());
			LibPlatform::Error::GetInstance()->Report(LibPlatform::Error::ERROR_CODE_WEBKIT_ASSET_LOAD_FAIL, L"", 0);
			break;

		case WebCore::DomainAccessControl::DOMAIN_ACCESS_NOT_ENCRYPTED:
			OS::Report("This request must be over HTTPS. Please use the HTTPS protocol.\n\tRequested URL: %s.\n",
					 encodeWithURLEscapeSequences(url.string()).latin1().data());
			LOG(Network, "This request must be over HTTPS. Please use the HTTPS protocol.\n\tRequested URL: %s.\n",
				encodeWithURLEscapeSequences(url.string()).latin1().data());
			LibPlatform::Error::GetInstance()->Report(LibPlatform::Error::ERROR_CODE_CURLE_UNSUPPORTED_PROTOCOL, L"", 0);
			break;

		case WebCore::DomainAccessControl::DOMAIN_ACCESS_NOT_WHITELISTED:
			OS::Report("The domain '%s' is not whitelisted. Add it to the Whitelist page of Project Settings.\n",
					 encodeWithURLEscapeSequences(url.string()).latin1().data());
			LOG(Network, "The domain '%s' is not whitelisted. Add it to the Whitelist page of Project Settings.\n",
				encodeWithURLEscapeSequences(url.string()).latin1().data());
			LibPlatform::Error::GetInstance()->Report(LibPlatform::Error::ERROR_CODE_WEBKIT_ADDRESS_BLOCKED_BY_WHITELIST, L"", 0);
			break;

		case WebCore::DomainAccessControl::DOMAIN_ACCESS_NOT_ACQUIRED_AUTH_TOKEN:
			OS::Report("Before accessing a network resource, an independent service token must first be acquired.\n" \
					"Please call nwf.act.NintendoAccountManager.acquireNewIndependentServiceToken().\n");
			LOG(Network, "Before accessing a network resource, an independent service token must first be acquired.\n" \
					"Please call nwf.act.NintendoAccountManager.acquireNewIndependentServiceToken().\n");
			break;

		default:
			OS::Report("This domain is unreachable. Error %d\n", error);
			LOG(Network, "This domain is unreachable. Error %d\n", error);
			break;
		}

        return 0;
	}
#endif

    // create our Curl handle
    m_CURLhandle = curl_easy_init();
    // If we can't get a CURL handle quit.
    if ( !m_CURLhandle ) {
        LOG(Network, "Unable to create CURL handle, quitting\n");
        return 0;
    }
    m_port = urlcopy.port();
    // curl doesn't understand a ws:// or wss:// protocol, so switch it to http
    if ( urlcopy.protocolIs("ws") || urlcopy.protocolIs("wss")) {
        urlcopy.setProtocol("http");
    }
    m_url = urlcopy.string();

    // Curl connect and setup
    curl_easy_setopt(m_CURLhandle, CURLOPT_URL,  m_url.latin1().data());
    curl_easy_setopt(m_CURLhandle, CURLOPT_PORT, m_port);
    curl_easy_setopt(m_CURLhandle, CURLOPT_CONNECT_ONLY, 1L);
    curl_easy_setopt(m_CURLhandle, CURLOPT_CONNECTTIMEOUT, 5); // we'll block for up to 5 seconds waiting for the connection to be made.
	curl_easy_setopt(m_CURLhandle, CURLOPT_SSL_CIPHER_LIST, LibPlatform::SSL::GetCipherList());
	if (*LibPlatform::Network::GetCertificatePath())
		curl_easy_setopt(m_CURLhandle, CURLOPT_CAINFO, LibPlatform::Network::GetCertificatePath());
	if (*LibPlatform::Network::GetClientCertificatePath())
		curl_easy_setopt(m_CURLhandle, CURLOPT_SSLCERT, LibPlatform::Network::GetClientCertificatePath());

    rc = curl_easy_perform(m_CURLhandle);
      while ( rc != CURLE_OK ) { 
		rc = curl_easy_perform(m_CURLhandle);
		retryCount++; 
		if(retryCount >= 10) {
			LOG(Network, "Error : CURL failed to Connect to URL (%s) with error : %s", m_url.latin1().data(), curl_easy_strerror(rc));
			//m_streamHandle->client()->didFail(m_streamHandle, SocketStreamError(rc, m_url.latin1().data()));
			return 0;
		}
    } 
	LOG(Network, "Created CURL link to WebSocket, marking our handle as open\n");
	if ( m_streamHandle && m_streamHandle->client() ) {
		// setup the socket
		if ( !getSocketInfo() ) {
			LOG(Network, "Unable to aquire socket, WebSocket functionality disabled");
			//m_streamHandle->client()->didFail(m_streamHandle, SocketStreamError(-1, m_url.latin1().data()));
			return 0;
		}
		m_streamHandle->m_state =  SocketStreamHandleBase::Open;
		// put a callback on the timer queue to mark channel as open.  Required since WebSocketChannel needs
		// reference to this SocketStreamHandle before call to didOpen
		setup_timer.startOneShot(0);
	}

    // check the WebSocket for data every 100ms
    m_timer.startRepeating(0.1);
    return 1;
}

SocketStreamHandle::SocketStreamHandle(const KURL& url, SocketStreamHandleClient* client)
    : SocketStreamHandleBase(url, client)
{
    int rc = 0;
    m_p = new WebCore::SocketStreamHandlePrivate(this);
    rc = m_p->initialize(url);

    LOG(Network, "%s creation of SocketStreamHandler for WebSocket", rc == 1 ? "Succesful" : "Failed" );

}

SocketStreamHandle::~SocketStreamHandle()
{
    setClient(0);
    delete m_p;
}

int SocketStreamHandle::platformSend(const char* buf, int length)
{
    LOG(Network, "SocketStreamHandle %p platformSend\n", this);
    return m_p->send(buf, length);
}

void SocketStreamHandle::platformClose()
{
    LOG(Network, "SocketStreamHandle %p platformClose\n", this);
    if ( m_p->m_CURLhandle ) {
        curl_easy_cleanup(m_p->m_CURLhandle);
        m_p->m_CURLhandle = 0;
    }
}

void SocketStreamHandle::didReceiveAuthenticationChallenge(const AuthenticationChallenge&)
{
    LOG(Network, "SocketStreamHandle %p didReceiveAuthenticationChallenge\n", this);
    notImplemented();
}

void SocketStreamHandle::receivedCredential(const AuthenticationChallenge&, const Credential&)
{
    LOG(Network, "SocketStreamHandle %p receivedCredential\n", this);
    notImplemented();
}

void SocketStreamHandle::receivedRequestToContinueWithoutCredential(const AuthenticationChallenge&)
{
    LOG(Network, "SocketStreamHandle %p receivedRequestToContinueWithoutCredential\n", this);
    notImplemented();
}

void SocketStreamHandle::receivedCancellation(const AuthenticationChallenge&)
{
    LOG(Network, "SocketStreamHandle %p recivedCancellation\n", this);
    notImplemented();
}

}  // namespace WebCore
