/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef WiiUEvent_h
#define WiiUEvent_h

#include <config.h>
#include <wtf/Noncopyable.h>
#include <wtf/PassOwnPtr.h>

#if PLATFORM(WIIU)
#include <sys/socket.h> // htonl, htons, ntohl, ntohs
#else
#if OS(WINDOWS)
#include <winsock2.h> // htonl, htons, ntohl, ntohs
#else
#include <arpa/inet.h> // htonl, htons, ntohl, ntohs
#endif
#endif

// -------------------------------------------------------------------------------------

typedef int	WiiUEventType_t;
typedef void * WebViewHandle_t;

// -------------------------------------------------------------------------------------

inline void hton_WebViewHandle(WebViewHandle_t handle, unsigned char *buffer)
{
	for(unsigned int i = 0; i < (sizeof(WebViewHandle_t) / sizeof(unsigned int)); ++i)
		((unsigned int *)buffer)[i] = htonl(((unsigned int *)&handle)[i]);
}
inline void ntoh_WebViewHandle(WebViewHandle_t *handle, unsigned char *buffer)
{
	for(unsigned int i = 0; i < (sizeof(WebViewHandle_t) / sizeof(unsigned int)); ++i)
		((unsigned int *)handle)[i] = ntohl(((unsigned int *)buffer)[i]);
}

// -------------------------------------------------------------------------------------

//! Event IDs used within WebCore
enum {
	WIIU_EMPTY_EVENT = 0,
	WIIU_TIMER_FIRED,
	WIIU_REDRAW,
	WIIU_RUN_ON_MAIN,
	WIIU_MOUSE_DOWN,
	WIIU_MOUSE_MOVE,
	WIIU_MOUSE_UP,
	WIIU_TOUCH_DOWN,
	WIIU_TOUCH_MOVE,
	WIIU_TOUCH_UP,
	WIIU_KEY_DOWN,
	WIIU_KEY_UP,

	WIIU_USER_EVENT_START,
};

enum {
    WIIU_MAX_DISPLAY = 3,            //!< Maximum number of display indices supported by WiiUDisplayVsyncEvent -> 1xTV, 2xGamePad
};

// -------------------------------------------------------------------------------------
// Event data (not used directly in event queue, used with EventData definitions below)
//

class WiiUMouseEvent
{
public:
	WiiUMouseEvent(int _x, int _y, int _button, int _state, int _clickCount, long long _timestamp) : x(_x), y(_y), button(_button), state(_state), clickCount(_clickCount), timestamp(_timestamp) {}
	WiiUMouseEvent(unsigned char *buffer) { fromSerialBuffer(buffer); }
	
	int x;
	int y;
	int button;
	int state;
	int clickCount;
	long timestamp;

	int serialSize(void) const { return sizeof(int)*4; }
	void toSerialBuffer(unsigned char *buffer) const
	{
		*((int *)(buffer                )) = htonl(x);
		*((int *)(buffer + sizeof(int)  )) = htonl(y);
		*((int *)(buffer + sizeof(int)*2)) = htonl(button);
		*((int *)(buffer + sizeof(int)*3)) = htonl(state);
		*((int *)(buffer + sizeof(int)*4)) = htonl(clickCount);
		*((long long *)(buffer + sizeof(int)*5)) = htonl(timestamp);
	}
	void fromSerialBuffer(unsigned char *buffer)
	{
		x      = ntohl(*((int *)(buffer                )));
		y      = ntohl(*((int *)(buffer + sizeof(int)  )));
		button = ntohl(*((int *)(buffer + sizeof(int)*2)));
		state  = ntohl(*((int *)(buffer + sizeof(int)*3)));
		clickCount = ntohl(*((int *)(buffer + sizeof(int)*4)));
		timestamp = ntohl(*((long long *)(buffer + sizeof(int)*5)));
	}
};

class WiiUKeyboardEvent
{
public:
	WiiUKeyboardEvent(int _state, int _scancode, int _sym, int _mod, unsigned short _unicode)
		: state(_state), scancode(_scancode), sym(_sym), mod(_mod), unicode(_unicode) {}
	WiiUKeyboardEvent(unsigned char *buffer) { fromSerialBuffer(buffer); }

	int 			state;
	int 			scancode;
    int 			sym;
    int 			mod;
    unsigned short 	unicode;

	int serialSize(void) const { return sizeof(int)*4 + sizeof(unsigned short); }
	void toSerialBuffer(unsigned char *buffer) const
	{
		*((int *)           (buffer                )) = htonl(state);
		*((int *)           (buffer + sizeof(int)  )) = htonl(scancode);
		*((int *)           (buffer + sizeof(int)*2)) = htonl(sym);
		*((int *)           (buffer + sizeof(int)*3)) = htonl(mod);
		*((unsigned short *)(buffer + sizeof(int)*4)) = htons(unicode);
	}
	void fromSerialBuffer(unsigned char *buffer)
	{
		state    = ntohl(*((int *)           (buffer                )));
		scancode = ntohl(*((int *)           (buffer + sizeof(int)  )));
		sym      = ntohl(*((int *)           (buffer + sizeof(int)*2)));
		mod      = ntohl(*((int *)           (buffer + sizeof(int)*3)));
		unicode  = ntohs(*((unsigned short *)(buffer + sizeof(int)*4)));
	}
};

// -------------------------------------------------------------------------------------

//! base class for all event datas passed through the event queue
class WiiUBaseEventData
{
public:
	virtual ~WiiUBaseEventData() {}
	WiiUBaseEventData() {}
	
	virtual WiiUEventType_t	type() const		{ return WIIU_EMPTY_EVENT; }

	virtual bool rejectExclusive() const		{ return false; }
	virtual void postExclusive()				{ /* nothing to do here */ }
	virtual void releaseExclusive()				{ /* nothing to do here */ }

	#if 0 // FIXME?: Should be a pure-virtual type, to enforce unique implementations? -TS-NSA 2012-04-10
	virtual int  serialSize(void) const           = 0;
	virtual void toSerialBuffer(unsigned char *buffer) const = 0;
	virtual void fromSerialBuffer(unsigned char *buffer)     = 0;
	#else
	virtual int  serialSize(void) const           { return 0; }
	virtual void toSerialBuffer(unsigned char *buffer) const { }
	virtual void fromSerialBuffer(unsigned char *buffer)     { }
	#endif
};

// -------------------------------------------------------------------------------------
// WebCore Event Classes
//

class WiiURunOnMainEvent : public WiiUBaseEventData
{
public:
	virtual WiiUEventType_t	type() const	{ return WIIU_RUN_ON_MAIN; }
};


class WiiUTimerFired : public WiiUBaseEventData
{
public:
	WiiUTimerFired(void *_pFunction) : pFunction(_pFunction) {}
	
	void*	pFunction;
	
	virtual WiiUEventType_t	type() const	{ return WIIU_TIMER_FIRED; }

	virtual bool rejectExclusive() const;
	virtual void postExclusive();
	virtual void releaseExclusive();

	// TODO?: this event isn't currently serializable -TS-NSA 2012-04-10

	friend void ResetWkEventGlobals();
	static void Startup(void);
private:
	static bool		s_bInQueue;
};


class WiiURedrawEvent : public WiiUBaseEventData
{
public:
	WiiURedrawEvent(WebViewHandle_t _webview) : webview(_webview) {}
	WiiURedrawEvent(unsigned char *buffer) { fromSerialBuffer(buffer); }
	
	WebViewHandle_t		webview;

	virtual WiiUEventType_t	type() const	{ return WIIU_REDRAW; }

	virtual int serialSize(void) const { return sizeof(WebViewHandle_t); }
	virtual void toSerialBuffer(unsigned char *buffer) const
	{
		hton_WebViewHandle(webview, buffer);
	}
	virtual void fromSerialBuffer(unsigned char *buffer)
	{
		ntoh_WebViewHandle(&webview, buffer);
	}
};

// ------------------------------------------------------------------------------------------------

class WiiUKeyboardData : public WiiUBaseEventData
{
public:
	WiiUKeyboardData(WiiUEventType_t _evType, WebViewHandle_t _webview, int _state, int _scancode, int _sym, int _mod, unsigned short _unicode)
		: ev(_state, _scancode, _sym, _mod, _unicode), webview(_webview), evType(_evType) {}
	WiiUKeyboardData(unsigned char *buffer) : ev(0, 0, 0, 0, 0) { fromSerialBuffer(buffer); }

	WiiUKeyboardEvent	ev;
	WebViewHandle_t		webview;
	int					evType;

	virtual WiiUEventType_t	type() const	{ return (WiiUEventType_t)evType; }

	virtual int serialSize(void) const { return ev.serialSize() + sizeof(WebViewHandle_t) + sizeof(int); }
	virtual void toSerialBuffer(unsigned char *buffer) const
	{
		ev.toSerialBuffer(buffer);
		hton_WebViewHandle(webview, buffer + ev.serialSize());
		*((int *)(buffer + ev.serialSize() + sizeof(WebViewHandle_t))) = htonl(evType);
	}
	virtual void fromSerialBuffer(unsigned char *buffer)
	{
		ev.fromSerialBuffer(buffer);
		ntoh_WebViewHandle(&webview, buffer + ev.serialSize());
		evType = ntohl(*((int *)(buffer + ev.serialSize() + sizeof(WebViewHandle_t))));
	}
};

class WiiUMouseData : public WiiUBaseEventData
{
public:
	WiiUMouseData(WiiUEventType_t _evType, WebViewHandle_t _webview, int _x, int _y, int _button, int _state, int _clickCount = 0, long long _timestamp = 0)
		: ev(_x, _y, _button, _state, _clickCount, _timestamp), webview(_webview), evType(_evType) {}
	WiiUMouseData(unsigned char *buffer) : ev(0, 0, 0, 0, 0, 0) { fromSerialBuffer(buffer); }
	WiiUMouseData() : ev(0, 0, 0, 0, 0, 0) {}
	
	WiiUMouseEvent		ev;
	WebViewHandle_t		webview;
	int					evType;

	virtual WiiUEventType_t	type() const	{ return (WiiUEventType_t)evType; }

    virtual bool rejectExclusive() const;
    virtual void postExclusive();
    virtual void releaseExclusive();
	friend void ResetWkEventGlobals();
    static void Startup(void);

	virtual int serialSize(void) const { return ev.serialSize() + sizeof(WebViewHandle_t) + sizeof(int); }
	virtual void toSerialBuffer(unsigned char *buffer) const
	{
		ev.toSerialBuffer(buffer);
		hton_WebViewHandle(webview, buffer + ev.serialSize());
		*((int *)(buffer + ev.serialSize() + sizeof(WebViewHandle_t))) = htonl(evType);
	}
	virtual void fromSerialBuffer(unsigned char *buffer)
	{
		ev.fromSerialBuffer(buffer);
		ntoh_WebViewHandle(&webview, buffer + ev.serialSize());
		evType = ntohl(*((int *)(buffer + ev.serialSize() + sizeof(WebViewHandle_t))));
	}


private:
    static bool           s_bMouseMoveInQueue[WIIU_MAX_DISPLAY];
    static WiiUMouseData * s_LastMouseMoveEvent[WIIU_MAX_DISPLAY];

};

// -------------------------------------------------------------------------------------

namespace WebCore {

// -------------------------------------------------------------------------------------

//! Event class to harbor event data in event queue
class WiiUEvent {
public:
	static PassOwnPtr<WiiUEvent> create(WiiUBaseEventData* data);
	WiiUEvent(WiiUBaseEventData* data) : m_data(data) {}
	~WiiUEvent()					{ if(m_data) delete m_data; }

	WiiUBaseEventData *getData()	{ return m_data; }
	WiiUBaseEventData *ownData()	{ WiiUBaseEventData *data = m_data; m_data = NULL; return data; }

private:
	WiiUBaseEventData *m_data;
};

} // namespace WebCore

#endif
