/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include "WiiUEventManager.h"
#include "NullPtr.h"

#include <wtf/CurrentTime.h>

using namespace WebCore;

// Global instance of the event manager.
WiiUEventManager *WiiUEventManager::m_instance = NULL;

/* Constructor for the event manager
 */
WiiUEventManager::WiiUEventManager()
{
	pthread_mutex_init(&m_queueMutex, NULL);
}

/*
 * Tear down the event manager.  This will invoke the m_eventQueue destructor.
 */
WiiUEventManager::~WiiUEventManager()
{
	pthread_mutex_destroy(&m_queueMutex);
}

WiiUEventManager *WiiUEventManager::instance()
{
	if (m_instance == NULL)
		m_instance = new WiiUEventManager();
	return m_instance;
}

namespace WebCore
{
	void ResetWiiUEventManagerGlobals()
	{
		WiiUEventManager::destroy();
	}
}

/*
 * Called when you want to destory the event manager.
 */
void WiiUEventManager::destroy()
{
	if(m_instance)
		delete m_instance;
	m_instance = NULL;
}

/*
 * Puts an event in to the queue.  This will cause anything that is blocked on the event manager waiting
 * for an event to unblock.
 */
void WiiUEventManager::putEvent(WiiUBaseEventData* data)
{
	pthread_mutex_lock(&m_queueMutex);

	if (data->rejectExclusive())
	{
		pthread_mutex_unlock(&m_queueMutex);

		delete data;
		return;
	}

	data->postExclusive();
	OwnPtr<WiiUEvent> ev = WiiUEvent::create(data);
	m_eventQueue.append(ev.release()); 

	pthread_mutex_unlock(&m_queueMutex);
}

/**
	Wait for event to arrive

	@param	data	Pointer to pointer to receive event data

	@note	Updated data reference will receive the oldest event in the queue
*/
void WiiUEventManager::waitEvent(WiiUBaseEventData **data)
{
	OwnPtr<WiiUEvent> ev = m_eventQueue.waitForMessage();
	*data = ev->ownData();

	pthread_mutex_lock(&m_queueMutex);
	(*data)->releaseExclusive();
	pthread_mutex_unlock(&m_queueMutex);
}

class DummyPredicate {
public:
	bool operator()(const WiiUEvent *) const { return true; }
};

/**
	Wait for event to arrive or timeout to expire

	@param	data	Pointer to pointer to receive event data

	@return			True if data was retrieved, false if timeout occured

	@note	Updated data reference will receive the oldest event in the queue
*/
bool WiiUEventManager::waitEventWithTimeout(WiiUBaseEventData **data, double timeoutSeconds)
{
	DummyPredicate dummyPredicate;
	MessageQueueWaitResult res;
	double absTime = timeoutSeconds + WTF::currentTime();
	OwnPtr<WiiUEvent> ev = m_eventQueue.waitForMessageFilteredWithTimeout(res, dummyPredicate, absTime);

	if (res != MessageQueueMessageReceived)
		return false;

	*data = ev->ownData();

	pthread_mutex_lock(&m_queueMutex);
	(*data)->releaseExclusive();
	pthread_mutex_unlock(&m_queueMutex);

	return true;
}

int WiiUEventManager::getNextEvent(WiiUBaseEventData **data)
{
	OwnPtr<WiiUEvent> ev = m_eventQueue.tryGetMessage();

	if(!ev || !ev.get())
		return 0;

	*data = ev->ownData();

	pthread_mutex_lock(&m_queueMutex);
	(*data)->releaseExclusive();
	pthread_mutex_unlock(&m_queueMutex);

	return 1;
}

